import functools
import operator
from typing import List

import pyarrow.compute as pc

from src.platform.common.errors import InvalidInputError
from src.platform.common.internal_dataclasses import AndFilterExpr, FilterType


def convert_filter_expr_to_pyarrow_compute_filter(
    filter_exprs: List[AndFilterExpr], table_columns: List[str]
) -> pc.Expression:
    """Converts filter expressions and table column into 1 filter expression for offboarding.

    Args:
        filter_exprs: List of filter expressions to offboard data for
        table_columns: List of columns from the table

    Returns:
        combined AND expression of all filter expressions.
    """
    filters = []
    table_columns_set = set(list(table_columns))

    for filter_expr in filter_exprs:
        if filter_expr.column not in table_columns_set:
            continue

        if filter_expr.filter_type == FilterType.EQUAL:
            filters.append(
                (
                    pc.not_equal(
                        pc.field(filter_expr.column), pc.scalar(filter_expr.value)
                    )
                )
            )
        elif filter_expr.filter_type == FilterType.EQUAL_IN:
            if not isinstance(filter_expr.value, list):
                raise InvalidInputError("Value must be list for: EQUAL_IN filter type")
            else:
                array_filters = []
                for scalar_value in list(filter_expr.value):
                    array_filters.append(
                        (
                            pc.not_equal(
                                pc.field(filter_expr.column), pc.scalar(scalar_value)
                            )
                        )
                    )
                filters.append(functools.reduce(operator.or_, array_filters))
        elif filter_expr.filter_type == FilterType.GREATER:
            filters.append(
                ~(
                    pc.greater(
                        pc.field(filter_expr.column), pc.scalar(filter_expr.value)
                    )
                )
            )
        elif filter_expr.filter_type == FilterType.GREATER_EQUAL:
            filters.append(
                ~(
                    pc.greater_equal(
                        pc.field(filter_expr.column), pc.scalar(filter_expr.value)
                    )
                )
            )
        elif filter_expr.filter_type == FilterType.LESSER:
            filters.append(
                ~(pc.less(pc.field(filter_expr.column), pc.scalar(filter_expr.value)))
            )
        elif filter_expr.filter_type == FilterType.LESSER_EQUAL:
            filters.append(
                ~(
                    pc.less_equal(
                        pc.field(filter_expr.column), pc.scalar(filter_expr.value)
                    )
                )
            )

    if filters:
        # combine the filters . this is always AND expression
        return functools.reduce(operator.and_, filters)
    else:
        return None
