from typing import List

from src.platform.offboarder.file_handlers import (
    FileMetadata,
    PyArrowTableWrapper,
    FileSystemHandlerInterface,
)
from src.platform.common.filter import convert_filter_expr_to_pyarrow_compute_filter
from concurrent.futures import ProcessPoolExecutor


"""
NOTE: time to offboard 1GB, 10K files 

No-multi threading:  35s 
ThreadPoolExecutor:  8s with 10 CPU cores
ProcessPoolExecutor: 8s with 10 CPU cores

Paralleization Factor: 4.375/10 per core 

m7g.16xlarge gives you 64 vCPUs. costs $2.6/hour.
    -> theorectically would finish this task in 1second. 35 / (0.4375 * 64) = $1.25
    approx 1GB/sec.
    In that hour, you can get 3.6TB cleaned.
    With adjusting for S3 latency. assume 1TB per hour.

    To clean 200TB of data, you only need: $520 dollars

"""


class DataOffboarder:
    def __init__(
        self,
        max_workers: int,
        filter_exprs,
        fs_handler: FileSystemHandlerInterface,
    ):
        self._max_workers = max_workers
        self._filter_exprs = filter_exprs
        self._fs_handler = fs_handler

    def offboard_data_lambda(self, file_metadata: FileMetadata):
        orig_table_wrapper = self._fs_handler.read_table(file_metadata)

        orig_table = orig_table_wrapper.table
        column_names = orig_table.schema.names
        offboard_filter_expr = convert_filter_expr_to_pyarrow_compute_filter(
            self._filter_exprs, column_names
        )

        filtered_table = orig_table.filter(offboard_filter_expr)
        filtered_table_wrapper = PyArrowTableWrapper(
            table=filtered_table, file_metadata=orig_table_wrapper.file_metadata
        )
        if len(filtered_table) < len(orig_table):
            print(f"Ovewriting: {file_metadata.path}")
            self._fs_handler.overwrite(filtered_table_wrapper, file_metadata)
        elif len(filtered_table) == len(orig_table):
            print(f"{file_metadata.path} had no change")
            return
        else:
            # case that filtered_table is bigger than original.
            # thereotically never possible because we run `filter` which is strictly less or equal to before
            raise ValueError(
                "Aborted. Somehow offboarded data size is now greater than original"
            )

    def __call__(self):
        file_metadatas_to_offboard: List[
            FileMetadata
        ] = self._fs_handler.iterate_file_metadata()

        with ProcessPoolExecutor(max_workers=self._max_workers) as executor:
            futures = [
                executor.submit(
                    self.offboard_data_lambda,
                    file_metadata,
                )
                for file_metadata in file_metadatas_to_offboard
            ]
            for future in futures:
                # TODO: add error handling and logging
                future.result()
