from dataclasses import dataclass
import os
import tempfile
from typing import Callable, Dict

import boto3
from botocore.client import BaseClient
import pyarrow
import pyarrow.parquet as pa_parquet
from pyarrow.parquet import FileMetaData as PyArrowFileMetaData

from src.platform.common.constants import PARQUET_COMPRESSION_METADATA_TO_COMPRESSION


def get_file_extension(file_path):
    return file_path[file_path.rfind(".") :]


def get_s3_bucket_and_key(s3_path) -> str:
    return s3_path.replace("s3://", "").split("/", 1)


def is_supported_file_extension(file_path: str) -> bool:
    file_extension = get_file_extension(file_path)
    # TODO: support the following types:
    # - ORC
    # - JSON
    # - CSV
    # - AVRO
    return file_extension in (".parquet")


def is_hudi_parquet(schema):
    if schema.metadata:
        for metadata_key in schema.metadata.keys():
            if "hoodie" in str(metadata_key):
                return True
    return False


def is_iceberg_parquet(schema):
    if schema.metadata:
        for metadata_key in schema.metadata.keys():
            if "iceberg" in str(metadata_key):
                return True
    return False


def is_delta_lake_parquet(schema):
    # NOTE: cannot be done from reading the file's metadata
    # instead, the offboarder needs to check for exsitance of: `_delta_log` folder
    return False


@dataclass(frozen=True)
class FileMetadata:
    path: str
    size: int
    extension: str


@dataclass(frozen=True)
class PyArrowTableWrapper:
    table: pyarrow.Table
    file_metadata: PyArrowFileMetaData


class FileSystemHandlerInterface:
    def __init__(self):
        pass

    def read_table(self, file: FileMetadata) -> PyArrowTableWrapper:
        raise NotImplementedError("read_table not implemented")

    def overwrite(self, table: pyarrow.Table, file: FileMetadata):
        raise NotImplementedError("overwrite not implemented")

    def validate_folder(self):
        raise NotImplementedError("validate_folder not implemented")


class LocalFileSystemHandler(FileSystemHandlerInterface):
    def __init__(self, folder_path: str):
        self._folder_path = folder_path

    def get_file_metadata(self, file_path: str) -> FileMetadata:
        file_size = os.stat(file_path).st_size

        return FileMetadata(
            path=file_path, size=file_size, extension=get_file_extension(file_path)
        )

    def iterate_file_metadata(self):
        for root, _, files in os.walk(self._folder_path):
            for file_path in files:
                full_path = os.path.join(root, file_path)
                if is_supported_file_extension(full_path):
                    yield self.get_file_metadata(full_path)

    def read_table(self, file: FileMetadata) -> PyArrowTableWrapper:
        return PyArrowTableWrapper(
            table=_read_table_local(file),
            file_metadata=pa_parquet.read_metadata(file.path),
        )

    def overwrite(self, table: pyarrow.Table, file: FileMetadata) -> bool:
        return _overwrite_local(table, file)

    def validate_folder(self):
        if not (os.path.exists(self._folder_path) and os.path.isdir(self._folder_path)):
            raise ValueError(f"Specified {self._folder_path} does not exist")


class S3FileSystemHandler(FileSystemHandlerInterface):
    def __init__(
        self,
        folder_path,
        s3_extra_args: Dict = None,
        s3_generator_callable: Callable[[Dict], BaseClient] = None,
    ):
        self._folder_path = folder_path
        self._s3_extra_args = s3_extra_args
        # s3 client can be optionally instanitated outside this class
        # if specific settings need to be applied
        self._s3_client = (
            s3_generator_callable(s3_extra_args)
            if s3_generator_callable
            else self._default_s3_generator_callable(s3_extra_args)
        )

    def iterate_file_metadata(self):
        # Split the S3 prefix into bucket and prefix components
        s3_parts = get_s3_bucket_and_key(self._folder_path)
        bucket_name = s3_parts[0]
        prefix = s3_parts[1] if len(s3_parts) > 1 else ""

        # Initialize the list of keys

        # Use the S3 client to list objects in the specified bucket and prefix,
        # handling pagination as necessary
        paginator = self._s3_client.get_paginator("list_objects_v2")
        response_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
        for response in response_iterator:
            if "Contents" in response:
                for obj in response["Contents"]:
                    # add file filter here
                    s3_key = f's3://{bucket_name}/{obj["Key"]}'
                    if is_supported_file_extension(s3_key):
                        yield FileMetadata(
                            path=s3_key,
                            size=obj["Size"],
                            extension=get_file_extension(s3_key),
                        )

    def read_table(self, file: FileMetadata) -> pyarrow.Table:
        s3_path = file.path

        bucket_name, file_key = get_s3_bucket_and_key(s3_path)

        with tempfile.NamedTemporaryFile(suffix=".parquet", delete=True) as tmp_file:
            self._s3_client.download_file(
                Bucket=bucket_name, Key=file_key, Filename=tmp_file.name
            )
            table = _read_table_local(
                FileMetadata(
                    path=tmp_file.name, size=file.size, extension=file.extension
                )
            )
            return PyArrowTableWrapper(
                table=table, file_metadata=pa_parquet.read_metadata(tmp_file.name)
            )

    def overwrite(self, table_wrapper: PyArrowTableWrapper, file: FileMetadata) -> bool:
        s3_path = file.path
        bucket_name, file_key = get_s3_bucket_and_key(s3_path)
        with tempfile.NamedTemporaryFile(suffix=".parquet", delete=True) as tmp_file:
            _overwrite_local(
                table_wrapper,
                FileMetadata(
                    path=tmp_file.name, size=file.size, extension=file.extension
                ),
            )
            self._s3_client.upload_file(
                tmp_file.name,
                bucket_name,
                file_key,
                ExtraArgs=self._s3_extra_args if self._s3_extra_args else None,
            )
        return True

    def validate_folder(self):
        bucket_name, _ = get_s3_bucket_and_key(self._folder_path)
        response = self._s3_client.get_bucket_versioning(Bucket=bucket_name)
        is_bucketing_on = "Status" in response and response["Status"] == "Enabled"

        if not is_bucketing_on:
            raise ValueError("Bucketing must be enabled")

    def _default_s3_generator_callable(self, s3_extra_args: Dict) -> BaseClient:
        if s3_extra_args:
            return boto3.client("s3", **s3_extra_args)
        return boto3.client("s3")


def _read_table_local(file: FileMetadata) -> pyarrow.Table:
    if file.extension == ".parquet":
        return pa_parquet.read_table(file.path)
    elif file.extension == ".orc":
        raise ValueError("orc not supported but planned in future release")
    elif file.extension == ".csv":
        raise ValueError("csv not supported but planned in future release")
    elif file.extension == ".json":
        raise ValueError("JSON not supported but planned in future release")
    else:
        raise ValueError("Not supported type")


def _overwrite_local(table_wrapper: PyArrowFileMetaData, file: FileMetadata) -> bool:
    if file.extension == ".parquet":
        # must infer the compression format first
        compression_metadata = (
            table_wrapper.file_metadata.row_group(0).column(0).compression
        )
        pa_parquet.write_table(
            table_wrapper.table,
            file.path,
            compression=PARQUET_COMPRESSION_METADATA_TO_COMPRESSION[
                compression_metadata
            ],
        )
    elif file.extension == ".orc":
        raise ValueError("orc not supported but planned in future release")
    elif file.extension == ".csv":
        raise ValueError("csv not supported but planned in future release")
    elif file.extension == ".json":
        raise ValueError("JSON not supported but planned in future release")
    else:
        raise ValueError("Not supported type")
    return True
