from dataclasses import dataclass
import io
import tempfile
import time
from typing import List, Iterable

import boto3
import pyarrow
import pyarrow.parquet as pq

from src.platform.common.internal_dataclasses import AndFilterExpr, OffboardLog
from src.platform.common.filter import convert_filter_expr_to_pyarrow_compute_filter
from src.platform.common.timer import Timer
from src.platform.common.logger import LOG, SUMMARY_LOGGER


@dataclass
class S3Metadata:
    s3_path: str
    size: int

    @property
    def data_id(self):
        return self.s3_path


class ParquetOffboarder:
    _YES_PROMPT = "y"

    def __init__(self):
        self._s3_client = boto3.client("s3")

    def _is_bucket_versioning_enabled(self, bucket_name: str):
        response = self._s3_client.get_bucket_versioning(Bucket=bucket_name)
        # cache the response
        return "Status" in response and response["Status"] == "Enabled"

    def _abort_no_versioning(self):
        raise Exception("Aborted by user because versioning is not enabled")

    def _warning_input(self):
        LOG.info("WARNING: Detected bucket without versioning enabled!!")
        user_prompt_warn = input(
            "WARNING: THIS IS NOT RECOMMENDED FOR BUCKETS WITHOUT VERSIONING ENABLED. \n Press `y` to accept this risk: "
        )
        if user_prompt_warn != ParquetOffboarder._YES_PROMPT:
            self._abort_no_versioning()
        user_prompt_final_warn = input(
            "WARNING: ARE YOU SURE? \n Press `y`, this is final warning: "
        )
        if user_prompt_final_warn != ParquetOffboarder._YES_PROMPT:
            self._abort_no_versioning()

    def __call__(
        self,
        filter_exprs: List[AndFilterExpr],
        s3_prefix: str,
        is_dry_run: True,
        requires_prompt=True,
    ):
        """
        Args:
            filter_exprs:       list of filter expression. treated as ANDs
            s3_prefix:          s3 prefix in which to look for .parquet files
                                and delete based on filter expr
            is_dry_run:         whether it's a dry_run. where it will not delete any data.
                                True by default for safety reasons
            requires_prompt:    boolean to require `y` answer on `input` for every file clean-up
                                True by default for safety reasons
        """
        bucket_name, _ = get_s3_bucket_and_key(s3_prefix)

        if not self._is_bucket_versioning_enabled(bucket_name):
            self._warning_input()

        for s3_file_metadata in list_s3_files(self._s3_client, s3_prefix):
            s3_path = s3_file_metadata.s3_path

            has_executed_delete = False
            LOG.info(f"\t Reading: {s3_path}")

            with Timer() as timer:
                orig_table = get_parquet_table_from_s3(self._s3_client, s3_path)
                column_names: List[str] = orig_table.schema.names
                offboard_filter_expr = convert_filter_expr_to_pyarrow_compute_filter(
                    filter_exprs, column_names
                )

                if offboard_filter_expr is None:
                    LOG.info(f"\t\t no applicable filters for {s3_path}. skipping")
                    continue

                filtered_table = orig_table.filter(offboard_filter_expr)
                LOG.info(
                    f"\t\t before: {len(orig_table)}, after: {len(filtered_table)}"
                )
                LOG.info(f"\t\t\t applied filters: {offboard_filter_expr}")

                if len(filtered_table) == 0:
                    # TODO: open question: if this was the case that all tables content got deleted,
                    # should we delete the whole file?
                    pass
                elif len(filtered_table) > len(orig_table):
                    # thereotically never possible because we run `filter` which makes things less or equal to before
                    raise ValueError(
                        "Aborted. Somehow offboarded data size is now greater than original"
                    )
                elif len(orig_table) != len(filtered_table):
                    rows_diff = len(orig_table) - len(filtered_table)

                    if not is_dry_run:
                        if requires_prompt:
                            user_input = input(
                                f"\t\t # of rows to be deleted: {rows_diff} \n\t\t to over-write \t: {s3_path} \n\t\t press `y` \t: "
                            ).strip()
                            if user_input != ParquetOffboarder._YES_PROMPT:
                                LOG.info(
                                    f"\t\t Skipped {s3_path} due to user requesting to skip it"
                                )

                        LOG.info(
                            f"\t\t writing cleaned data of new size: {len(filtered_table)}, to: {s3_path}"
                        )
                        has_executed_delete = write_parquet_table_to_s3(
                            filtered_table, self._s3_client, s3_path
                        )
                    else:
                        LOG.info(
                            f"\t\t dry-run stats: {len(orig_table)} to {len(filtered_table)} rows for: {s3_path} "
                        )

                SUMMARY_LOGGER.info(
                    OffboardLog(
                        data_id=s3_file_metadata.data_id,
                        before_count=len(orig_table),
                        after_count=len(filtered_table)
                        if not is_dry_run
                        else len(orig_table),
                        has_executed_delete=has_executed_delete,
                        offboaded_epoch=time.time(),
                    )
                )
            if has_executed_delete:
                LOG.info(f"\t Elapsed: {round(timer.elapsed_sec, 2)}s for: {s3_path}")


# utils
def get_s3_bucket_and_key(s3_path) -> str:
    return s3_path.replace("s3://", "").split("/", 1)


def get_parquet_table_from_s3(s3_client, s3_path) -> pyarrow.Table:
    bucket_name, file_key = get_s3_bucket_and_key(s3_path)
    buffer = io.BytesIO()
    s3_client.download_fileobj(Bucket=bucket_name, Key=file_key, Fileobj=buffer)
    return pq.read_table(buffer)


def write_parquet_table_to_s3(table, s3_client, s3_path, s3_extra_args=None) -> bool:
    bucket_name, file_key = get_s3_bucket_and_key(s3_path)
    with tempfile.NamedTemporaryFile(suffix=".parquet") as tmp_file:
        pq.write_table(table, tmp_file.name)
        s3_client.upload_file(
            tmp_file.name,
            bucket_name,
            file_key,
            ExtraArgs=s3_extra_args if s3_extra_args else None,
        )
    return True


def list_s3_files(s3_client, s3_prefix) -> Iterable[str]:
    # Split the S3 prefix into bucket and prefix components
    s3_parts = get_s3_bucket_and_key(s3_prefix)
    bucket_name = s3_parts[0]
    prefix = s3_parts[1] if len(s3_parts) > 1 else ""

    # Initialize the list of keys

    # Use the S3 client to list objects in the specified bucket and prefix,
    # handling pagination as necessary
    paginator = s3_client.get_paginator("list_objects_v2")
    response_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    for response in response_iterator:
        if "Contents" in response:
            for obj in response["Contents"]:
                if is_filter_parquet_file(obj["Key"]):
                    yield S3Metadata(
                        s3_path=f's3://{bucket_name}/{obj["Key"]}', size=obj["Size"]
                    )


def is_filter_parquet_file(file_path):
    return file_path.endswith(".parquet")
