import io
import os
import tempfile
import unittest


from src.platform.benchmarking.mock_parquet_generator import MockParquetGenerator
from src.platform.common.internal_dataclasses import AndFilterExpr, FilterType
from src.platform.offboarder.data_offboarder import DataOffboarder, FileMetadata
from src.platform.offboarder.file_handlers import (
    LocalFileSystemHandler,
    S3FileSystemHandler,
)

import boto3
from botocore.stub import Stubber
from botocore import stub

import pyarrow
import pyarrow.compute as pc
import pyarrow.parquet as pa_parquet


class TestDataOffboarder(unittest.TestCase):
    def setUp(self) -> None:
        self._mock_parquet_generator = MockParquetGenerator()
        self._s3_client = boto3.client("s3")

    def test_data_offboarder_local(self):
        temp_dir = tempfile.mkdtemp()

        with tempfile.NamedTemporaryFile(
            dir=temp_dir, suffix=".parquet", delete=True
        ) as tmp_file:
            mock_pa_table: pyarrow.Table = (
                self._mock_parquet_generator.generate_sample_pa_table(100)
            )
            pa_parquet.write_table(mock_pa_table, tmp_file.name)

            data_offboarder = DataOffboarder(
                max_workers=os.cpu_count(),
                filter_exprs=[
                    AndFilterExpr(
                        filter_type=FilterType.EQUAL, column="mod_ten", value="1"
                    ),
                ],
                fs_handler=LocalFileSystemHandler(temp_dir),
            )
            data_offboarder()

            after_offboarding: pyarrow.Table = pa_parquet.read_table(tmp_file.name)
            self.assertEqual(len(after_offboarding), 90)
            self.assertEqual(
                after_offboarding,
                mock_pa_table.filter(~pc.equal(pc.field("mod_ten"), pc.scalar("1"))),
            )

    def test_data_offboarder_lambda_local(self):
        temp_dir = tempfile.mkdtemp()

        # delete case
        with tempfile.NamedTemporaryFile(
            dir=temp_dir, suffix=".parquet", delete=True
        ) as tmp_file:
            mock_pa_table: pyarrow.Table = (
                self._mock_parquet_generator.generate_sample_pa_table(100)
            )
            pa_parquet.write_table(mock_pa_table, tmp_file.name)

            data_offboarder = DataOffboarder(
                max_workers=os.cpu_count(),
                filter_exprs=[
                    AndFilterExpr(
                        filter_type=FilterType.EQUAL, column="mod_ten", value="1"
                    ),
                ],
                fs_handler=LocalFileSystemHandler(temp_dir),
            )
            data_offboarder.offboard_data_lambda(
                FileMetadata(path=tmp_file.name, size=-1, extension=".parquet")
            )

            after_offboarding: pyarrow.Table = pa_parquet.read_table(tmp_file.name)
            self.assertEqual(len(after_offboarding), 90)
            self.assertEqual(
                after_offboarding,
                mock_pa_table.filter(~pc.equal(pc.field("mod_ten"), pc.scalar("1"))),
            )

        # equal case
        with tempfile.NamedTemporaryFile(
            dir=temp_dir, suffix=".parquet", delete=True
        ) as tmp_file:
            mock_pa_table: pyarrow.Table = (
                self._mock_parquet_generator.generate_sample_pa_table(100)
            )
            pa_parquet.write_table(mock_pa_table, tmp_file.name)

            data_offboarder = DataOffboarder(
                max_workers=os.cpu_count(),
                filter_exprs=[
                    AndFilterExpr(
                        filter_type=FilterType.EQUAL, column="mod_ten", value="billion"
                    ),
                ],
                fs_handler=LocalFileSystemHandler(temp_dir),
            )
            data_offboarder.offboard_data_lambda(
                FileMetadata(path=tmp_file.name, size=-1, extension=".parquet")
            )

            after_offboarding: pyarrow.Table = pa_parquet.read_table(tmp_file.name)
            self.assertEqual(len(after_offboarding), 100)
            self.assertEqual(after_offboarding, mock_pa_table)

    def test_data_offboard_lambda_s3(self):
        mock_pa_table: pyarrow.Table = (
            self._mock_parquet_generator.generate_sample_pa_table(100)
        )
        data_offboarder = DataOffboarder(
            max_workers=os.cpu_count(),
            filter_exprs=[
                AndFilterExpr(
                    filter_type=FilterType.EQUAL, column="mod_ten", value="billion"
                ),
            ],
            fs_handler=S3FileSystemHandler(
                "s3://mock-bucket/mock_prefix",
                s3_generator_callable=lambda s3_extra_args: self._s3_client,
            ),
        )

        with Stubber(self._s3_client) as stubber:
            io_buffer = None
            with tempfile.NamedTemporaryFile(suffix=".parquet") as tmp_file:
                pa_parquet.write_table(mock_pa_table, tmp_file.name)
                with open(tmp_file.name, "rb") as file:
                    io_buffer = io.BytesIO(file.read())

            stubber.add_response(
                "list_objects_v2",
                {
                    "Contents": [
                        {"Key": "mock_prefix/mock_prefix/file1.parquet", "Size": 12345},
                    ]
                },
                {"Bucket": "mock-bucket", "Prefix": "mock_prefix"},
            )

            head_or_get_obj_param = {
                "Bucket": "mock-bucket",
                "Key": "mock_prefix/mock_prefix/file1.parquet",
            }
            stubber.add_response(
                "head_object",
                {
                    "ContentLength": 10,
                    "ContentType": "utf-8",
                    "ResponseMetadata": {
                        "Bucket": "mock-bucket",
                    },
                },
                head_or_get_obj_param,
            )
            stubber.add_response(
                "get_object",
                {
                    "ContentLength": 10,
                    "ContentType": "utf-8",
                    "Body": io_buffer,
                    "ResponseMetadata": {
                        "Bucket": "mock-bucket",
                    },
                },
                head_or_get_obj_param,
            )
            stubber.add_response(
                "put_object",
                {},
                {
                    "Body": stub.ANY,
                    "Bucket": "mock-bucket",
                    "Key": "mock_pa_table.parquet",
                },
            )

            for _ in data_offboarder._fs_handler.iterate_file_metadata():
                print("\t", _)

            data_offboarder.offboard_data_lambda(
                FileMetadata(
                    path="s3://mock-bucket/mock_prefix/mock_prefix/file1.parquet",
                    size=12345,
                    extension=".parquet",
                )
            )


if __name__ == "__main__":
    unittest.main()
