import io
import tempfile
import unittest

import boto3
from botocore.stub import Stubber
from botocore import stub

import pyarrow.parquet as pa_parquet

from src.platform.benchmarking.mock_parquet_generator import MockParquetGenerator
from src.platform.offboarder.file_handlers import (
    FileMetadata,
    LocalFileSystemHandler,
    S3FileSystemHandler,
    FileSystemHandlerInterface,
    get_file_extension,
    is_supported_file_extension,
    is_hudi_parquet,
    is_iceberg_parquet,
)


class TestLocalFileSystemHandler(unittest.TestCase):
    def setUp(self):
        self._temp_dir = tempfile.mkdtemp()
        self._local_fs_handler = LocalFileSystemHandler(self._temp_dir)
        self._s3_client = boto3.client("s3")
        self._s3_fs_handler = S3FileSystemHandler(
            "s3://mock-bucket/mock_prefix",
            s3_generator_callable=lambda s3_extra_args: self._s3_client,
        )
        self._mock_parquet_generator = MockParquetGenerator()

    def test_parquet_compression_preserving(self):
        mock_table = self._mock_parquet_generator.generate_sample_pa_table(1)
        for compression_scheme in {"snappy", "gzip", "brotli", "zstd", "lz4", "none"}:
            with tempfile.NamedTemporaryFile(suffix=".parquet") as tmp_file:
                mock_file_metadata = FileMetadata(
                    path=tmp_file.name, size=0, extension=".parquet"
                )
                pa_parquet.write_table(
                    mock_table, mock_file_metadata.path, compression=compression_scheme
                )

                ret_before = self._local_fs_handler.read_table(mock_file_metadata)
                self._local_fs_handler.overwrite(ret_before, mock_file_metadata)
                ret_after = self._local_fs_handler.read_table(mock_file_metadata)

                self.assertEqual(
                    ret_before.file_metadata.row_group(0).column(0).compression,
                    ret_after.file_metadata.row_group(0).column(0).compression,
                )

                self.assertEqual(ret_before.table, ret_after.table)

    def test_get_file_extension(self):
        self.assertEqual(get_file_extension("mock.parquet"), ".parquet")
        self.assertEqual(get_file_extension("mock..parquet"), ".parquet")
        self.assertEqual(get_file_extension("mock.txt"), ".txt")

    def test_is_supported_file_extension(self):
        self.assertTrue(is_supported_file_extension("hi.parquet"))

    def test_is_hudi_parquet(self):
        pass
        # TODO: check-in mock hudi data
        # schema = pa_parquet.read_schema('../../ab961f27-0023-41c3-a710-acb4cc8b84ce-0_167-25-2235_20230529222810468.parquet')
        # self.assertTrue(is_hudi_parquet(schema))

    def test_interface(self):
        fs_handler_interface_instance = FileSystemHandlerInterface()
        with self.assertRaises(NotImplementedError):
            fs_handler_interface_instance.overwrite(
                None, FileMetadata(path="mock.parquet", size=0, extension=".parquet")
            )
        with self.assertRaises(NotImplementedError):
            fs_handler_interface_instance.read_table(
                FileMetadata(path="mock.parquet", size=0, extension=".parquet")
            )
        with self.assertRaises(NotImplementedError):
            fs_handler_interface_instance.validate_folder()

    def test_local_fs_handler(self):
        num_mock_parquet_files = 10
        for i in range(0, num_mock_parquet_files):
            mpg_res = self._mock_parquet_generator.generate_sample_pa_table(
                100, 100 * i
            )
            with tempfile.NamedTemporaryFile(
                dir=self._temp_dir, suffix=".parquet", delete=False
            ) as tmp_file:
                pa_parquet.write_table(mpg_res, tmp_file.name)

        self.assertEqual(
            len(list(self._local_fs_handler.iterate_file_metadata())),
            num_mock_parquet_files,
        )

    def test_local_fs_handler_validate(self):
        with self.assertRaises(ValueError):
            LocalFileSystemHandler("non_existant_folder").validate_folder()

    def test_s3_fs_handler_iterate_file_metadata(self):
        with Stubber(self._s3_client) as stubber:
            expected_params = {"Bucket": "mock-bucket", "Prefix": "mock_prefix"}
            mock_s3_file_responses = [
                {"Key": "mock_prefix/file1.parquet", "Size": 12345},
                {"Key": "mock_prefix/file2.parquet", "Size": 67890},
                {"Key": "mock_prefix/file3.parquet", "Size": 24680},
                {"Key": "mock_prefix/file4.parquet", "Size": 13579},
                {"Key": "mock_prefix/file5.parquet", "Size": 97531},
            ]
            expected_response = {"Contents": mock_s3_file_responses}
            stubber.add_response("list_objects_v2", expected_response, expected_params)

            self.assertEqual(
                list(self._s3_fs_handler.iterate_file_metadata()),
                [
                    FileMetadata(
                        path=f"s3://mock-bucket/{response['Key']}",
                        size=response["Size"],
                        extension=".parquet",
                    )
                    for response in mock_s3_file_responses
                ],
            )

    def test_s3_fs_handler_read_table_and_overwrite(self):
        mock_pa_table = self._mock_parquet_generator.generate_sample_pa_table(100)

        with Stubber(self._s3_client) as stubber:
            io_buffer = None
            with tempfile.NamedTemporaryFile(suffix=".parquet") as tmp_file:
                pa_parquet.write_table(mock_pa_table, tmp_file.name)
                with open(tmp_file.name, "rb") as file:
                    io_buffer = io.BytesIO(file.read())

            head_or_get_obj_param = {
                "Bucket": "mock-bucket",
                "Key": "mock_pa_table.parquet",
            }
            stubber.add_response(
                "head_object",
                {
                    "ContentLength": 10,
                    "ContentType": "utf-8",
                    "ResponseMetadata": {
                        "Bucket": "mock-bucket",
                    },
                },
                head_or_get_obj_param,
            )
            stubber.add_response(
                "get_object",
                {
                    "ContentLength": 10,
                    "ContentType": "utf-8",
                    "Body": io_buffer,
                    "ResponseMetadata": {
                        "Bucket": "mock-bucket",
                    },
                },
                head_or_get_obj_param,
            )
            stubber.add_response(
                "put_object",
                {},
                {
                    "Body": stub.ANY,
                    "Bucket": "mock-bucket",
                    "Key": "mock_pa_table.parquet",
                },
            )

            read_table_ret_val = self._s3_fs_handler.read_table(
                FileMetadata(
                    path="s3://mock-bucket/mock_pa_table.parquet",
                    size=0,
                    extension=".parquet",
                )
            )
            self.assertEqual(
                read_table_ret_val.table,
                mock_pa_table,
            )

            self._s3_fs_handler.overwrite(
                read_table_ret_val,
                FileMetadata(
                    path="s3://mock-bucket/mock_pa_table.parquet",
                    size=0,
                    extension=".parquet",
                ),
            )

    def test_s3_fs_handler_validate_folder(self):
        with Stubber(self._s3_client) as stubber:
            expected_params = {"Bucket": "mock-bucket"}
            expected_response = {"Status": "Enabled", "MFADelete": "Disabled"}

            # Stub the S3 API call
            stubber.add_response(
                "get_bucket_versioning", expected_response, expected_params
            )
            self._s3_fs_handler.validate_folder()

        # raises case: Suspended
        with Stubber(self._s3_client) as stubber:
            expected_params = {"Bucket": "mock-bucket"}
            expected_response = {"Status": "Suspended", "MFADelete": "Disabled"}
            # Stub the S3 API call
            stubber.add_response(
                "get_bucket_versioning", expected_response, expected_params
            )
            with self.assertRaises(ValueError):
                self._s3_fs_handler.validate_folder()

        # raises case: Empty
        with Stubber(self._s3_client) as stubber:
            expected_params = {"Bucket": "mock-bucket"}
            expected_response = {"Status": "", "MFADelete": "Disabled"}
            # Stub the S3 API call
            stubber.add_response(
                "get_bucket_versioning", expected_response, expected_params
            )
            with self.assertRaises(ValueError):
                self._s3_fs_handler.validate_folder()

    def test_unsupported_file_formats(self):
        unsupported_file_formats = {".json", ".csv", ".orc"}
        for unsupported_file_format in unsupported_file_formats:
            with tempfile.NamedTemporaryFile(
                suffix=f"{unsupported_file_format}"
            ) as tmp_file:
                with self.assertRaises(ValueError):
                    self._local_fs_handler.read_table(
                        FileMetadata(
                            path=tmp_file.name,
                            size=0,
                            extension=unsupported_file_format,
                        )
                    )

    def test_is_hudi_parquet(self):
        table = pa_parquet.read_table("src/testdata/userdata1.parquet")
        self.assertFalse(is_hudi_parquet(table.schema))
        self.assertTrue(
            is_hudi_parquet(
                # https://github.com/apache/hudi/blob/master/hudi-common/src/main/java/org/apache/hudi/common/model/HoodiePartitionMetadata.java#L58
                table.schema.with_metadata(
                    {".hoodie_partition_metadata": "mock_prefix"}
                )
            )
        )

    def test_is_iceberg_parquet(self):
        iceberg_table = pa_parquet.read_table(
            "src/testdata/iceberg_samples/iceberg_sample1.parquet"
        )
        self.assertTrue(is_iceberg_parquet(iceberg_table.schema))

        standard_table = pa_parquet.read_table("src/testdata/userdata1.parquet")
        self.assertFalse(is_iceberg_parquet(standard_table.schema))

    def test_is_delta_lake_parquet(self):
        pa_parquet.read_table(
            "src/testdata/delta_samples/part-00000-374e3521-a0fb-479b-89bc-d5f874f0b42d-c000.snappy.parquet"
        )
        pa_parquet.read_metadata(
            "src/testdata/delta_samples/part-00000-374e3521-a0fb-479b-89bc-d5f874f0b42d-c000.snappy.parquet"
        )


if __name__ == "__main__":
    unittest.main()
