import unittest

import pyarrow as pa

from src.platform.common.errors import InvalidInputError
from src.platform.common.filter import convert_filter_expr_to_pyarrow_compute_filter
from src.platform.common.internal_dataclasses import AndFilterExpr, FilterType


_MOCK_SCHEMA = pa.schema(
    [
        pa.field("str_col1", pa.string()),
        pa.field("int_col1", pa.int64()),
    ]
)
_COLUMNS = ["str_col1", "int_col1"]
_MOCK_ROW_1 = {"str_col1": "str_col1:val1", "int_col1": 1}
_MOCK_ROW_2 = {"str_col1": "str_col1:val2", "int_col1": 2}
_MOCK_ROW_3 = {"str_col1": "str_col1:val3", "int_col1": 3}

_MOCK_PYARROW_TABLE = pa.Table.from_pylist(
    [_MOCK_ROW_1, _MOCK_ROW_2, _MOCK_ROW_3],
    schema=_MOCK_SCHEMA,
)


class TestFilter(unittest.TestCase):
    def test_convert_filter_expr_to_pyarrow_compute_filter(self):
        equal_filter_expr = AndFilterExpr(
            filter_type=FilterType.EQUAL, column="str_col1", value="str_col1:val1"
        )
        equal_in_filter_expr_with_invalid_val = AndFilterExpr(
            filter_type=FilterType.EQUAL_IN, column="str_col1", value="invalid_val1"
        )
        equal_in_filter_expr_with_valid_val = AndFilterExpr(
            filter_type=FilterType.EQUAL_IN, column="str_col1", value=["str_col1:val1"]
        )
        greater_filter_expr = AndFilterExpr(
            filter_type=FilterType.GREATER, column="int_col1", value=1
        )
        greater_equal_filter_expr = AndFilterExpr(
            filter_type=FilterType.GREATER_EQUAL, column="int_col1", value=3
        )
        lesser_filter_expr = AndFilterExpr(
            filter_type=FilterType.LESSER, column="int_col1", value=2
        )
        lesser_equal_filter_expr = AndFilterExpr(
            filter_type=FilterType.LESSER_EQUAL, column="int_col1", value=2
        )

        self.assertEqual(
            convert_filter_expr_to_pyarrow_compute_filter([equal_filter_expr], []),
            None,
        )

        with self.assertRaises(InvalidInputError):
            convert_filter_expr_to_pyarrow_compute_filter(
                [equal_in_filter_expr_with_invalid_val], _COLUMNS
            )

        # EQUAL
        self.assertEqual(
            _MOCK_PYARROW_TABLE.filter(
                convert_filter_expr_to_pyarrow_compute_filter(
                    [equal_filter_expr], _COLUMNS
                )
            ).to_pylist(),
            [_MOCK_ROW_2, _MOCK_ROW_3],
        )

        # EQUAL_IN
        self.assertEqual(
            _MOCK_PYARROW_TABLE.filter(
                convert_filter_expr_to_pyarrow_compute_filter(
                    [equal_in_filter_expr_with_valid_val], _COLUMNS
                )
            ).to_pylist(),
            [_MOCK_ROW_2, _MOCK_ROW_3],
        )

        # GREATER
        self.assertEqual(
            _MOCK_PYARROW_TABLE.filter(
                convert_filter_expr_to_pyarrow_compute_filter(
                    [greater_filter_expr], _COLUMNS
                )
            ).to_pylist(),
            [_MOCK_ROW_1],
        )

        # GREATER_EQUAL
        self.assertEqual(
            _MOCK_PYARROW_TABLE.filter(
                convert_filter_expr_to_pyarrow_compute_filter(
                    [greater_equal_filter_expr], _COLUMNS
                )
            ).to_pylist(),
            [_MOCK_ROW_1, _MOCK_ROW_2],
        )

        # LESSER
        self.assertEqual(
            _MOCK_PYARROW_TABLE.filter(
                convert_filter_expr_to_pyarrow_compute_filter(
                    [lesser_filter_expr], _COLUMNS
                )
            ).to_pylist(),
            [_MOCK_ROW_2, _MOCK_ROW_3],
        )

        # LESSER_EQUAL
        self.assertEqual(
            _MOCK_PYARROW_TABLE.filter(
                convert_filter_expr_to_pyarrow_compute_filter(
                    [lesser_equal_filter_expr], _COLUMNS
                )
            ).to_pylist(),
            [_MOCK_ROW_3],
        )


if __name__ == "__main__":
    unittest.main()
